@@ -17,11 +17,14 @@
  * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <meta/util.h>
 #include <meta/meta-background.h>
 #include <meta/meta-background-image.h>
 #include "meta-background-private.h"
 #include "cogl-utils.h"
 
+#include <string.h>
+
 enum
 {
   CHANGED,
@@ -36,7 +39,7 @@ struct _MetaBackgroundMonitor
 {
   gboolean dirty;
   CoglTexture *texture;
-  CoglOffscreen *fbo;
+  CoglFramebuffer *fbo;
 };
 
 struct _MetaBackgroundPrivate
@@ -71,6 +74,8 @@ enum
 
 G_DEFINE_TYPE (MetaBackground, meta_background, G_TYPE_OBJECT)
 
+static gboolean texture_has_alpha (CoglTexture *texture);
+
 static GSList *all_backgrounds = NULL;
 
 static void
@@ -312,6 +317,18 @@ meta_background_finalize (GObject *object)
   G_OBJECT_CLASS (meta_background_parent_class)->finalize (object);
 }
 
+static void
+meta_background_constructed (GObject *object)
+{
+  MetaBackground        *self = META_BACKGROUND (object);
+  MetaBackgroundPrivate *priv = self->priv;
+
+  G_OBJECT_CLASS (meta_background_parent_class)->constructed (object);
+
+  g_signal_connect_object (meta_screen_get_display (priv->screen), "gl-video-memory-purged",
+                           G_CALLBACK (mark_changed), object, G_CONNECT_SWAPPED);
+}
+
 static void
 meta_background_class_init (MetaBackgroundClass *klass)
 {
@@ -322,6 +339,7 @@ meta_background_class_init (MetaBackgroundClass *klass)
 
   object_class->dispose = meta_background_dispose;
   object_class->finalize = meta_background_finalize;
+  object_class->constructed = meta_background_constructed;
   object_class->set_property = meta_background_set_property;
   object_class->get_property = meta_background_get_property;
 
@@ -474,7 +492,7 @@ get_texture_area (MetaBackground          *self,
     }
 }
 
-static void
+static gboolean
 draw_texture (MetaBackground        *self,
               CoglFramebuffer       *framebuffer,
               CoglPipeline          *pipeline,
@@ -483,6 +501,7 @@ draw_texture (MetaBackground        *self,
 {
   MetaBackgroundPrivate *priv = self->priv;
   cairo_rectangle_int_t texture_area;
+  gboolean bare_region_visible;
 
   get_texture_area (self, monitor_area, texture, &texture_area);
 
@@ -503,6 +522,9 @@ draw_texture (MetaBackground        *self,
                                                 - texture_area.y / (float)texture_area.height,
                                                 (monitor_area->width - texture_area.x) / (float)texture_area.width,
                                                 (monitor_area->height - texture_area.y) / (float)texture_area.height);
+
+      bare_region_visible = texture_has_alpha (texture);
+
       /* Draw just the texture */
       break;
     case G_DESKTOP_BACKGROUND_STYLE_CENTERED:
@@ -513,11 +535,16 @@ draw_texture (MetaBackground        *self,
                                                 texture_area.x + texture_area.width,
                                                 texture_area.y + texture_area.height,
                                                 0, 0, 1.0, 1.0);
+      bare_region_visible = texture_has_alpha (texture) || memcmp (&texture_area, monitor_area, sizeof (cairo_rectangle_int_t)) != 0;
+      break;
     case G_DESKTOP_BACKGROUND_STYLE_NONE:
+      bare_region_visible = TRUE;
       break;
     default:
-      g_return_if_reached();
+      g_return_val_if_reached(FALSE);
     }
+
+  return bare_region_visible;
 }
 
 static void
@@ -529,6 +556,7 @@ ensure_color_texture (MetaBackground *self)
     {
       ClutterBackend *backend = clutter_get_default_backend ();
       CoglContext *ctx = clutter_backend_get_cogl_context (backend);
+      CoglError *error = NULL;
       uint8_t pixels[6];
       int width, height;
 
@@ -569,7 +597,13 @@ ensure_color_texture (MetaBackground *self)
                                                                          COGL_PIXEL_FORMAT_RGB_888,
                                                                          width * 3,
                                                                          pixels,
-                                                                         NULL));
+                                                                         &error));
+
+      if (error != NULL)
+        {
+          meta_warning ("Failed to allocate color texture: %s\n", error->message);
+          cogl_error_free (error);
+        }
     }
 }
 
@@ -628,6 +662,7 @@ ensure_wallpaper_texture (MetaBackground *self,
     {
       int width = cogl_texture_get_width (texture);
       int height = cogl_texture_get_height (texture);
+      CoglOffscreen *offscreen;
       CoglFramebuffer *fbo;
       CoglError *catch_error = NULL;
       CoglPipeline *pipeline;
@@ -635,7 +670,8 @@ ensure_wallpaper_texture (MetaBackground *self,
       priv->wallpaper_texture = meta_create_texture (width, height,
                                                      COGL_TEXTURE_COMPONENTS_RGBA,
                                                      META_TEXTURE_FLAGS_NONE);
-      fbo = cogl_offscreen_new_with_texture (priv->wallpaper_texture);
+      offscreen = cogl_offscreen_new_with_texture (priv->wallpaper_texture);
+      fbo = COGL_FRAMEBUFFER (offscreen);
 
       if (!cogl_framebuffer_allocate (fbo, &catch_error))
         {
@@ -748,13 +784,17 @@ meta_background_get_texture (MetaBackground         *self,
   if (monitor->dirty)
     {
       CoglError *catch_error = NULL;
+      gboolean bare_region_visible = FALSE;
 
       if (monitor->texture == NULL)
         {
+          CoglOffscreen *offscreen;
+
           monitor->texture = meta_create_texture (monitor_area.width, monitor_area.height,
                                                   COGL_TEXTURE_COMPONENTS_RGBA,
                                                   META_TEXTURE_FLAGS_NONE);
-          monitor->fbo = cogl_offscreen_new_with_texture (monitor->texture);
+          offscreen = cogl_offscreen_new_with_texture (monitor->texture);
+          monitor->fbo = COGL_FRAMEBUFFER (offscreen);
         }
 
       if (!cogl_framebuffer_allocate (monitor->fbo, &catch_error))
@@ -783,9 +823,9 @@ meta_background_get_texture (MetaBackground         *self,
           cogl_pipeline_set_layer_texture (pipeline, 0, texture2);
           cogl_pipeline_set_layer_wrap_mode (pipeline, 0, get_wrap_mode (priv->style));
 
-          draw_texture (self,
-                        monitor->fbo, pipeline,
-                        texture2, &monitor_area);
+          bare_region_visible = draw_texture (self,
+                                              monitor->fbo, pipeline,
+                                              texture2, &monitor_area);
 
           cogl_object_unref (pipeline);
         }
@@ -796,8 +836,7 @@ meta_background_get_texture (MetaBackground         *self,
                                     0.0, 0.0, 0.0, 0.0);
         }
 
-      if (texture1 != NULL &&
-          !(texture2 != NULL && priv->blend_factor == 1.0 && !texture_has_alpha (texture2)))
+      if (texture1 != NULL && priv->blend_factor != 1.0)
         {
           CoglPipeline *pipeline = create_pipeline (PIPELINE_ADD);
           cogl_pipeline_set_color4f (pipeline,
@@ -808,15 +847,14 @@ meta_background_get_texture (MetaBackground         *self,
           cogl_pipeline_set_layer_texture (pipeline, 0, texture1);
           cogl_pipeline_set_layer_wrap_mode (pipeline, 0, get_wrap_mode (priv->style));
 
-          draw_texture (self,
-                        monitor->fbo, pipeline,
-                        texture1, &monitor_area);
+          bare_region_visible = bare_region_visible || draw_texture (self,
+                                                                     monitor->fbo, pipeline,
+                                                                     texture1, &monitor_area);
 
           cogl_object_unref (pipeline);
         }
 
-      if (!((texture2 != NULL && priv->blend_factor == 1.0 && !texture_has_alpha (texture2)) ||
-            (texture1 != NULL && !texture_has_alpha (texture1))))
+      if (bare_region_visible)
         {
           CoglPipeline *pipeline = create_pipeline (PIPELINE_OVER_REVERSE);
 
